/*
 * Process-related convenience functions.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>
#include <notifier/notifier.h>
#include <systypes/env.h>

#include <stdio.h>
#include <stdlib.h>

#include "process_creator_context_client.h"

#include "file.h"
#include "process.h"



/* Create a new process object. */

process_t *process_new()
{
  process_t *process = (process_t *) malloc(sizeof(process_t));

  if (process == NULL)
    return NULL;

  process_init(process);
  return process;
}

/* Discard the given process. */

void process_close(process_t *process)
{
  if (process == NULL)
    return;

  if (l4_is_valid_cap(process->ref))
    ipc_cap_free_um(process->ref);

  process_init(process);
}

/* Return any process initiation error. */

long process_error(process_t *process)
{
  if (process->notifiable.notifications & NOTIFY_TASK_ALL)
    return process->notifiable.values.val;
  else
    return L4_EOK;
}

/* Discard and free the given process, similar to process_close but freeing the
   structure instead of reinitialising it. */

void process_free(process_t *process)
{
  if (process == NULL)
    return;

  if (l4_is_valid_cap(process->ref))
    ipc_cap_free_um(process->ref);

  free(process);
}

/* Initialise the given process structure. */

void process_init(process_t *process)
{
  process->ref = L4_INVALID_CAP;

  /* Initialise the notifiable section of the structure. */

  process->notifiable.notifications = 0;
  process->notifiable.pending_notifications = 0;
  process->notifiable.base = (notifiable_base_t *) process;
  process->notifiable.handler = NULL;
}

/* A convenience function for creating and starting a process. */

long process_spawn(int argc, const char *argv[], file_t *reader,
                   file_t *writer, file_t *error, process_t **process)
{
  *process = process_new();

  /* Start the process with the given arguments. */

  if (*process != NULL)
    return process_start(*process, argc, argv, reader, writer, error);
  else
    return -L4_ENOMEM;
}

/* Start a process using the given arguments. */

long process_start(process_t *process, int argc, const char *argv[],
                   file_t *reader, file_t *writer, file_t *error)
{
  l4_cap_idx_t server = l4re_env_get_cap(ENV_PROCESS_SERVER_NAME);

  if (l4_is_invalid_cap(server))
    return -L4_ENOENT;

  /* Obtain a context for process creation. */

  file_t context;
  long err = file_context(&context, server);

  if (err)
    return err;

  offset_t pos = 0, written;

  for (int i = 0; i < argc; i++)
  {
    if (!file_string_set(&context, argv[i], pos, &written))
      return -L4_ENOMEM;

    pos += written + 1;
  }

  /* Obtain a client for the process creator context. */

  client_ProcessCreatorContext creator(context.ref);

  /* Start the process, obtaining a reference to it. */

  err = creator.start(argc,
                      reader != NULL ? reader->ref : (l4_cap_idx_t) L4_INVALID_CAP,
                      writer != NULL ? writer->ref : (l4_cap_idx_t) L4_INVALID_CAP,
                      error != NULL ? error->ref : (l4_cap_idx_t) L4_INVALID_CAP,
                      &process->ref);

  /* Initialise the notifiable section of the structure. */

  process->notifiable.notifications = 0;
  process->notifiable.pending_notifications = 0;
  process->notifiable.base = (notifiable_base_t *) process;

  /* Close the context, although a separate mechanism could permit contexts to
     open several processes. */

  file_close(&context);
  return err;
}

/* A convenience function for waiting for a started process. */

long process_wait(process_t *process, notify_flags_t *flags, notify_values_t *values)
{
  /* Obtain the common notifier. */

  notifier_t *notifier = notify_get_task();

  /* Subscribe to the process for notifications. */

  long err = notify_subscribe(process_notifiable(process), NOTIFY_TASK_ALL, notifier);

  if (err)
    return err;

  /* Wait for a signal from the process. */

  err = process_notify_wait_process(process, notifier);

  if (err)
    return err;

  /* Obtain the notification flags and values. */

  *flags = process_notifications(process);
  *values = process_notification_values(process);

  /* Obtain any error, closing the process. */

  err = process_error(process);
  process_free(process);
  return err;
}



/* Conversion to the generic notification types. */

notifiable_t *process_notifiable(process_t *process)
{
  return notify_notifiable((notifiable_base_t *) process);
}

/* Return the notification flags for a process. */

notify_flags_t process_notifications(process_t *process)
{
  return notify_notifications((notifiable_base_t *) process);
}

/* Return the notification values for a process. */

notify_values_t process_notification_values(process_t *process)
{
  return notify_notification_values((notifiable_base_t *) process);
}

/* Return whether a notification has indicated process termination. */

int process_terminated(notifiable_t *notifiable)
{
  return ((notifiable->notifications & NOTIFY_TASK_SIGNAL) &&
          (notifiable->values.sig == 0)) ||
         (notifiable->notifications & NOTIFY_TASK_ERROR);
}

/* Wait for a notification event on a process. */

long process_notify_wait_process(process_t *process, notifier_t *notifier)
{
  long err = notify_wait(process_notifiable(process), notifier);

  /* Unsubscribe if a termination notification has been received. */

  if (!err && process_terminated(process_notifiable(process)))
    notify_unsubscribe(process_notifiable(process), notifier);

  return err;
}

/* Wait for notification events on processes. */

long process_notify_wait_processes(process_t **process, notifier_t *notifier)
{
  notifiable_t *notifiable;
  long err = notify_wait_many(&notifiable, notifier);

  /* Unsubscribe if a termination notification has been received. */

  if (!err && process_terminated(notifiable))
  {
    *process = (process_t *) notifiable->base;
    notify_unsubscribe(notifiable, notifier);
  }

  return err;
}

// vim: tabstop=2 expandtab shiftwidth=2
